// Description: adjusts the material properties of all nodes below it
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> shininess accepts values from 0.0 to 128.0 and all others from 0.0 to 1.0
//<li> the material property of all children nodes will be changed
//<li> use the cache option on child objects to avoid unwanted side effects
//</ul>
//
// Category: Attributes
// Author: Alex Hill
//           11/01/01
//           04/01/04 Alex Hill - added "alpha" message to adjust transparency
// Revision: 07/01/05 Alex Hill - fixed initial message bug and removed relative mode
//
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfString.h>
#include <Performer/pf/pfText.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pf/pfChannel.h>
#include "ygNetMessages.h"
#include "material.h"

#define CHANGE_AMBIENT    (1 << 0)
#define CHANGE_DIFFUSE    (1 << 1)
#define CHANGE_SPECULAR   (1 << 2)
#define CHANGE_EMISSION   (1 << 3)
#define CHANGE_SHININESS  (1 << 4)
#define CHANGE_ALPHA      (1 << 5)

extern "C" ygNode* construct_material(const char* name,bool master) { return new material(name,master); }

material::material(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("material",true);
    gSetList = new pfList;
    ambient.set(0,0,0);
    diffuse.set(0,0,0);
    emission.set(0,0,0);
    specular.set(0,0,0);
    alpha = 1.0;
    shininess = 0.0;
    changed = 0;
    addNetMessage("ambient",&ambient,YG_NET_VEC3);
    addNetMessage("diffuse",&diffuse,YG_NET_VEC3);
    addNetMessage("emission",&emission,YG_NET_VEC3);
    addNetMessage("specular",&specular,YG_NET_VEC3);
    addNetMessage("shininess",&shininess,YG_NET_FLOAT);
    addNetMessage("alpha",&alpha,YG_NET_FLOAT);
    addNetMessage("changed",&changed,YG_NET_INT);
    //dont include in documentation
    dontSaveNetMessage("changed");
    }

material::~material(void)
    {
    }

void material::reset(void)
    {
    //reset all colors and shininess to zero
    gSetList->reset();
    setGSetList(pfnode());
    ambient.set(0,0,0);
    diffuse.set(0,0,0);
    emission.set(0,0,0);
    specular.set(0,0,0);
    alpha = 1.0;
    shininess = 0.0;
    resetNetMessage("changed");
    ygNode::reset();
    } 

void material::message(const ygMessage& msg)
    {
    //set the ambient color of the material
    if (msg == "ambient")
        {
        setColor(GL_AMBIENT,ambient[0],ambient[1],ambient[2]);
        changed |= CHANGE_AMBIENT;
        }
    //set the diffuse color of the material
    else if (msg == "diffuse")
        {
        setColor(GL_DIFFUSE,diffuse[0],diffuse[1],diffuse[2]);
        changed |= CHANGE_DIFFUSE;
        }
    //set the emission color of the material
    else if (msg == "emission")
        {
        setColor(GL_EMISSION,emission[0],emission[1],emission[2]);
        changed |= CHANGE_EMISSION;
        }
    //set the specular color of the material
    else if (msg == "specular")
        {
        setColor(GL_SPECULAR,specular[0],specular[1],specular[2]);
        changed |= CHANGE_SPECULAR;
        }
    //set the shininess of the material
    else if (msg == "shininess")
        {
        setShininess(shininess);
        changed |= CHANGE_SHININESS;
        }
    //set the alpha of the material
    else if (msg == "alpha")
        {
        setAlpha(alpha);
        changed |= CHANGE_ALPHA;
        }
    //set the change status of the material
    else if (msg == "changed")
        {
        if (changed == 0)
            {
            gSetList->reset();
            setGSetList(pfnode());
            ambient.set(0,0,0);
            diffuse.set(0,0,0);
            emission.set(0,0,0);
            specular.set(0,0,0);
            alpha = 1.0;
            shininess = 0.0;
            }
        }
    else
        ygNode::message(msg);
    }
    
void material::app(void)
    {
    //if changed then set the geoset list
    if (isChanged(YG_CHANGE_GEOMETRY|YG_CHANGE_TREE))
        {
        gSetList->reset();
        setGSetList(pfnode());
        if (changed & CHANGE_AMBIENT)
            setColor(GL_AMBIENT,ambient[0],ambient[1],ambient[2]);
        if (changed & CHANGE_DIFFUSE)
            setColor(GL_DIFFUSE,diffuse[0],diffuse[1],diffuse[2]);
        if (changed & CHANGE_EMISSION)
            setColor(GL_EMISSION,emission[0],emission[1],emission[2]);
        if (changed & CHANGE_SPECULAR)
            setColor(GL_SPECULAR,specular[0],specular[1],specular[2]);
        if (changed & CHANGE_SHININESS)
            setShininess(shininess);
        if (changed & CHANGE_ALPHA)
            setAlpha(alpha);
        }
    ygNode::app();
    }
    
void material::setGSetList(pfNode *node)
    {
    int i;
    //if node type is a pfGeode
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode* geode = (pfGeode*)node;
        //for each geoset
        for (i=0; i < geode->getNumGSets(); i++)
            {
            //add geoset to the geoset list
            pfGeoSet* gSet = geode->getGSet(i);
            gSetList->add(gSet);
            }
        }
    //else, if node type is a pfText
    else if (node->isOfType(pfText::getClassType()))
        {
        pfText* text = (pfText*)node;
        //for each string
        for (i=0; i < text->getNumStrings(); i++)
            {
            //add geoset to the geoset list
            pfString* string = text->getString(i);
            pfGeoSet* gSet = new pfGeoSet;
            gSet->setGState(string->getGState());
            gSetList->add(gSet);
            }
        }
    //else, if node type is a pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup* group = (pfGroup*)node;
        //call setGSetList recursively on all children
        for (i=0; i < group->getNumChildren(); i++)
            setGSetList(group->getChild(i));
        }
    }

void material::setColor(int color,float r,float g,float b)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        pfGeoState* gState = gSet->getGState();
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        int frontMode,backMode;
        if (frontMtl)
            {
            frontMode = frontMtl->getColorMode(PFMTL_FRONT);
            if (frontMode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
                if (color == PFMTL_AMBIENT)
                    frontMode = PFMTL_CMODE_AMBIENT;
                else if (color == PFMTL_DIFFUSE)
                    frontMode = PFMTL_CMODE_DIFFUSE;
            if (color == frontMode)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    {
                    colors[i][0] = r;
                    colors[i][1] = g;
                    colors[i][2] = b;
                    }
                }
            else
                frontMtl->setColor(color,r,g,b);
            }
        if (backMtl && backMtl != frontMtl)
            {
            backMode = backMtl->getColorMode(PFMTL_BACK);
            if (backMode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
                if (color == PFMTL_AMBIENT)
                    backMode = PFMTL_CMODE_AMBIENT;
                else if (color == PFMTL_DIFFUSE)
                    backMode = PFMTL_CMODE_DIFFUSE;
            if (color == backMode)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    {
                    colors[i][0] = r;
                    colors[i][1] = g;
                    colors[i][2] = b;
                    }
                }
            else
                backMtl->setColor(color,r,g,b);
            }
        }
    }

void material::setShininess(float s)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        pfGeoState* gState = gSet->getGState();
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        if (frontMtl)
            frontMtl->setShininess(s);
        if (backMtl && backMtl != frontMtl)
            backMtl->setShininess(s);
        }
    }
   
void material::setAlpha(float a)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        gSet->setDrawBin(PFSORT_TRANSP_BIN);
        pfGeoState* gState = gSet->getGState();
        gState->setMode(PFSTATE_TRANSPARENCY, PFTR_BLEND_ALPHA);
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        int frontMode,backMode;
        if (frontMtl)
            {
            frontMode = frontMtl->getColorMode(PFMTL_FRONT);
            if (frontMode != PFMTL_CMODE_OFF)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    colors[i][3] = a;
                }
            else
                frontMtl->setAlpha(a);
            }
        if (backMtl && backMtl != frontMtl)
            {
            backMode = backMtl->getColorMode(PFMTL_BACK);
            if (backMode != PFMTL_CMODE_OFF)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    colors[i][3] = a;
                }
            else
                backMtl->setAlpha(a);
            }
        }
    }
