// Description: a base class for ygGeometry nodes
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> the wall mode is used to turn collision detection on
//<li> the floor mode is used to specify whether the model is treated as terrain
//<li> the draw mode is set to false, ygGeometry is not draw but collision detection can still be enabled
//</ul>
//
// Category: Geometry
// Author: Alex Hill
//           11/01/02
//           09/01/04 Alex Hill - moved into the core executable and added "cull" message
//           06/28/05 Alex Hill - cull and wireframe are only applied if changed from default
// Revision: 02/01/06 Alex Hill - added debug display geometry management
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGeode.h>
#include <iostream.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include <ygPFObjectCache.h>
#include <Performer/pfdu.h>
#include "ygGeometry.h"

extern "C" ygNode* construct_ygGeometry(const char* name,bool master) { return new ygGeometry(name,master); }

ygGeometry::ygGeometry(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygGeometry",true);
    isectMask = YG_ISECT_ALLMASK;
    floor = false;
    wall = false;
    draw = DRAW_ON;
    wireframe = false;
    cull = PFCF_BACK;
    object = new pfGroup;
    pfnode()->addChild(object);
    addNetMessage("floor",&floor,YG_NET_BOOL_NEGATE);
    addNetMessage("wall",&wall,YG_NET_BOOL_NEGATE);
    addNetMessage("draw",&draw,YG_NET_INT);
    addNetMessageMap("draw","false",DRAW_OFF);
    addNetMessageMap("draw","true",DRAW_ON);
    addNetMessageMap("draw","local",DRAW_LOCAL);
    addNetMessageMap("draw","remote",DRAW_REMOTE);
    addNetMessage("wireframe",&wireframe,YG_NET_BOOL_NEGATE);
    addNetMessage("cull",&cull,YG_NET_INT);
    addNetMessageMap("cull","none",PFCF_OFF);
    addNetMessageMap("cull","front",PFCF_FRONT);
    addNetMessageMap("cull","back",PFCF_BACK);
    addNetMessageMap("cull","both",PFCF_BOTH);
    }

ygGeometry::~ygGeometry(void)
    {
    }


void ygGeometry::reset(void)
    {
    //set floor to false
    resetNetMessage("floor");
    setFloor(floor);
    //set wall to false
    resetNetMessage("wall");
    setWall(wall);
    //set the draw flag to true
    resetNetMessage("draw");
    setDraw(draw);
    //set the wireframe flag to false
    if (wireframe)
        {
        resetNetMessage("wireframe");
        setWireframe(wireframe);
        }
    //set the cull mode to back faces
    if (cull != PFCF_BACK)
        {
        resetNetMessage("cull");
        setCull(cull);
        }
    ygNode::reset();
    }


void ygGeometry::message(const ygMessage& msg)
    {
    //set collision detection status
    if (msg == "wall")
        {
        setWall(wall);
        }
    //set walk navigation status
    else if (msg == "floor")
        {
        setFloor(floor);
        }
    //set the draw status
    else if (msg == "draw")
        {
        setDraw(draw);
        }
    //set wireframe mode
    else if (msg == "wireframe")
        {
        setWireframe(wireframe);
        }
    //set cull mode
    else if (msg == "cull")
        {
        setCull(cull);
        }
    else
        ygNode::message(msg);
    }


void ygGeometry::app(void)
    {
    if (isChanged(YG_CHANGE_GEOMETRY))
        {
        if (cull != PFCF_BACK)
            setCull(cull);
        if (wireframe)
            setWireframe(wireframe);
        setFloor(floor);
        setWall(wall);
        if (debugNode())
            {
            createDebugNode();
	    updateDebug();
            }
        }
    ygNode::app();
    }


void ygGeometry::setWall(bool val)
    {
    wall = val;
    if (val)
        isectMask |= YG_ISECT_WALLMASK;
    else
        isectMask &= ~YG_ISECT_WALLMASK;
    object->setTravMask(PFTRAV_ISECT,isectMask,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    }


void ygGeometry::setFloor(bool val)
    {
    floor = val;
    if (val)
        isectMask |= YG_ISECT_FLOORMASK;
    else
        isectMask &= ~YG_ISECT_FLOORMASK;
    object->setTravMask(PFTRAV_ISECT,isectMask,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    }


void ygGeometry::setDraw(int val)
    {
    draw = val;
    if (val == DRAW_ON || (isNetMaster() && val == DRAW_LOCAL) || (!isNetMaster() && val == DRAW_REMOTE))
        object->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET);
    else
        object->setTravMask(PFTRAV_DRAW,0,PFTRAV_SELF,PF_SET);
    }


void ygGeometry::setWireframe(bool val)
    {
    wireframe = val;
    if (val)
        setGeoState(PFSTATE_ENWIREFRAME,PF_ON,object);
    else
        setGeoState(PFSTATE_ENWIREFRAME,PF_OFF,object);
    }


void ygGeometry::setCull(int mode)
    {
    cull = mode;
    setGeoState(PFSTATE_CULLFACE,cull,object);
    }


void ygGeometry::setGeoState(uint64_t mode, int state, pfNode* node)
    {
    int i;
    //if node type is pfGeode then set geoset properties
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        //for each geode, call setGeosetTexture
        for (i=0; i < geode->getNumGSets(); i++)
            {
            pfGeoSet* gset = geode->getGSet(i);
            gset->getGState()->setMode(mode,state);
            }
        }
    //else, if node type is pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        //call this function recursively on all children
        for (i=0; i < group->getNumChildren(); i++)
            setGeoState(mode,state,group->getChild(i));
        }
    }


void ygGeometry::createDebugNode(void)
    {
    pfGeoSet* gset;
    pfGeode* geode = (pfGeode*)debugNode();
    if (geode)
        {
        while (geode->getNumGSets() > 0)
            {
            gset = geode->getGSet(0);
            if (gset)
                {
                geode->removeGSet(gset);
                pfDelete(gset);
                }
            }
        }
    else
        {
        geode = new pfGeode;
        if (!geode)
            {
            cerr << "ERROR: ygGeometry::makeDebugGeode: failed to create pfGeode\n";
            return;
            }
        setDebugNode((pfGroup*)geode);
        }
    makeDebugGeosets(geode);
    }


void ygGeometry::makeDebugGeosets(pfGeode* geode)
    {
    pfBox box;
    pfuTravCalcBBox(pfnode(),&box);
    pfGeoSet *gset;
    pfMatrix matrix;
    pfVec3 scale;
    if (box.min[0] < box.max[0])
        {
        scale[0] = 0.1*(box.max[0] - box.min[0]);
        scale[1] = 0.1*(box.max[1] - box.min[1]);
        scale[2] = 0.1*(box.max[2] - box.min[2]);
        }
    else
        {
        box.min.set(-0.25,-0.25,-0.25);
        box.max.set(0.25,0.25,0.25);
        scale.set(0.25,0.25,0.25);
        }
    //create corner 1
    gset = makeCornerGeoset();
    pfGeoState* gstate = new pfGeoState;
    gstate->setMode(PFSTATE_ENWIREFRAME,PF_ON);
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.min[0],box.min[1],box.min[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);
    //create corner 2   
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(90.0,0,0,1);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.max[0],box.min[1],box.min[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create corner 3  
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(180.0,0,0,1);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.max[0],box.max[1],box.min[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create corner 4  
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(-90.0,0,0,1);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.min[0],box.max[1],box.min[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create corner 5
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(-90.0,1,0,0);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.min[0],box.min[1],box.max[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create corner 6
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(180.0,0,1,0);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.max[0],box.min[1],box.max[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create corner 7
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(90.0,0,0,1);
    matrix.postRot(matrix,180.0,1,0,0);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.max[0],box.max[1],box.max[2]);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create corner 8
    gset = makeCornerGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeRot(180.0,1,0,0);
    pfdXformGSet(gset,matrix);
    matrix.makeScale(scale[0],scale[1],scale[2]);
    matrix.postTrans(matrix,box.min[0],box.max[1],box.max[2]);
    pfdXformGSet(gset,matrix);
    geode->setBound(NULL,PFBOUND_DYNAMIC);
    geode->addGSet(gset);
    }


pfGeoSet* ygGeometry::makeCornerGeoset(void)
    {
    pfGeoSet *gset = new pfGeoSet;
    gset->setPrimType(PFGS_TRIS);
    gset->setNumPrims(3);
    pfVec3 * verts = (pfVec3 *) pfMalloc(9 * sizeof(pfVec3),pfGetSharedArena());
    verts[0][0] = 0; verts[0][1] = 0; verts[0][2] = 0;
    verts[1][0] = 0; verts[1][1] = 1; verts[1][2] = 0;
    verts[2][0] = 1; verts[2][1] = 0; verts[2][2] = 0;
    verts[3][0] = 0; verts[3][1] = 0; verts[3][2] = 0;
    verts[4][0] = 1; verts[4][1] = 0; verts[4][2] = 0;
    verts[5][0] = 0; verts[5][1] = 0; verts[5][2] = 1;
    verts[6][0] = 0; verts[6][1] = 0; verts[6][2] = 0;
    verts[7][0] = 0; verts[7][1] = 0; verts[7][2] = 1;
    verts[8][0] = 0; verts[8][1] = 1; verts[8][2] = 0;
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    return gset;
    }
