// Description: loads and transforms models from any one of the accepted Performer model formats
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the aliases: object, staticObject
//<li> objects are an instance from a Performer object cache unless the cache mode is set false before the file message
//<li> Common file formats supported by Performer:
//<b> .pfb</b> - IRIS Performer fast binary format
//<b> .flt</b> - MultiGen OpenFlight format
//<b> .iv</b> - SGI Open Inventor format
//<b> .wrl</b> - VRML version 2.0 format
//<b> .obj</b> - Wavefront Technologies data format
//<b> .3ds</b> - AutoDesk 3DStudio binary data
//<b> .dxf</b> - AutoDesk AutoCAD ASCII format
//<li> The pfconv utility can be used to convert any format that Performer will read into the native Performer fast binary format(.pfb)
//<ol> 
//<li> type pvconv then the name of the file you want to convert, followed by the name of the converted file
//<li> <b>ex:</b> pfconv oldfile.flt newfile.pfb
//<li> see the man pages on pfconv for more information
//</ol>
//</ul>
//
// Category: Geometry
// Author: Dave Pape
//           11/01/01
//           10/15/02 Dave Pape - add optional cache flag to "file" msg, "reload" message, and included cache-state in net filename
//           09/01/04 Alex Hill - derived from ygGeometry class and added functionality to create a parent ygTransform when necessary
//           06/28/05 Alex Hill - added functionality to dereference uncached object attributes
// Revision: 02/01/06 Alex Hill - added setGeodeName funtion to set the name of each pfGeode to the node name for intersection picking
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pfdu.h>
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygPFObjectCache.h"
#include "ygNodeDB.h"
#include "ygObject.h"

extern "C" ygNode* construct_ygObject(const char* name,bool master) { return new ygObject(name,master); }

ygObject::ygObject(const char* name,bool master) : ygGeometry(name,master)
    {
    setClassName("ygObject",true);
    obj = NULL;
    transform = NULL;
    cache = true;
    curCache = true;
    addNetMessage("file",&filename_,YG_NET_FILE);
    addNetMessage("file",&curCache,YG_NET_BOOL);
    dontSendNetMessage("file");
    }


ygObject::~ygObject(void)
    {
    if (obj && curCache)
        object->removeChild(obj);
    }


void ygObject::reset(void)
    {
    //set cache to true
    cache = true;
    curCache = true;
    if (!transform && parent())
        {
        ygString transName(name());
        transName += "xform";
        if (parent()->name() == transName)
            transform = parent();
        }
    ygGeometry::reset();
    }


void ygObject::message(const ygMessage& msg)
    {
    //load the given filename with optional caching
    if (msg == "file")
        {
        loadFile(filename_);
        }
    //set the default object caching mode
    else if (msg == "cache")
        {
        if (msg.numArgs() > 0)
            cache = msg.boolArg(0);
        }
    //reload the currently loaded object
    else if (msg == "reload")
        {
        curCache = false;
        loadFile(filename_);
        }
    //set whether the object is drawn locally (deprectated)
    else if (msg == "showLocal")
        {
        if (msg.numArgs() > 0)
            {
            bool val = msg.boolArg(0);
            if(val)
                setDraw(DRAW_ON);
            else
                setDraw(DRAW_REMOTE);
            }
        else
            setDraw(DRAW_ON);
        }
    //create a parent ygTransform and pass the message
    else if (msg == "position" || msg == "orientation" || msg == "size" || msg == "updateInterval" || msg == "reliableKey")
        {
        createTransform();
        transform->parseMessage(msg);
        }
    else
        ygGeometry::message(msg);
    }


void ygObject::loadFile(const ygString& file)
    {
    //if object has been loaded then remove it as a child
    if (obj)
        {
        object->removeChild(obj);
        setChanged(YG_CHANGE_GEOMETRY | YG_CHANGE_TEXTURE);
        }
    if (file.length() > 0)
        {
        //if cache mode is true then retrieve from object cache
        if (!cache)
            curCache = false;
        if (curCache)
            obj = ygPFObjectCache::objectReference(file.c_str());
        //else, load the file using pfdLoadFile
        else
            {
            obj = pfdLoadFile(file.c_str());
            if (obj)
                ygPFObjectCache::makeUnique(obj);
            }
        //if loading was successful
        if (obj)
            {
            filename_ = file;
            //name pfGeode to ygObject name
            setGeodeName(obj);
            //add object as a child
            object->addChild(obj);
            netMessageChanged("file");
            obj->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
            //set intersection traversal mask
            obj->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK, PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE, PF_SET);
            setChanged(YG_CHANGE_GEOMETRY | YG_CHANGE_TEXTURE);
            }
        }
    else
        obj = NULL;
    }


void ygObject::createTransform(void)
    {
    if (!transform)
        {
        ygString transName(name());
        transName += "xform";
        ygNodeDB::createValidName("ygTransform",transName);
        transform = ygNodeDB::create("ygTransform",transName);
        parent()->addChild(transform);
        transform->addChild(this);
        transform->reset();
        }
    }


void ygObject::setGeodeName(pfNode* node)
    {
    //if node type is pfGeode then set geoset properties
    if (node->isOfType(pfGeode::getClassType()))
        node->setName(name().c_str());
    //else, if node type is pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        //call this function recursively on all children
        for (int i=0; i < group->getNumChildren(); i++)
            setGeodeName(group->getChild(i));
        }
    }


const ygString& ygObject::filename(void) const
    {
    return filename_;
    }


void ygObject::createDebugNode(void)
    {
    createTransform();
    ygGeometry::createDebugNode();
    }
